#
# Tu - Tcl Usefulness
#
# Stuart Cassoff
#
# Fall/Winter 2018
# Version 0.1
#
# Summer 2020
# Version 0.2
#

package require Tcl 8.5

namespace eval tu {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu
		title	Tu
		descr	{Tcl Usefulness}
		author	{Stuart Cassoff}
		when	{Summer 2020}
		version	0.2
	}]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2017,2018 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


variable defaultWindow .


# Tcl

proc randstr {{length {6}}} {
	set str ""
	for {set i 0} {$i < $length} {incr i} {
		append str [format %c [expr {97 + int(rand() * 26)}]]
	}
	return $str
}

# dict get with default
#
proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

# In My Namespace
#
proc imn {what} {
	return [string trimright [uplevel 1 {namespace current}] :]::$what
}

#
# String Cat
#
proc strcat {args} { join $args "" }

#
# [lmap] replacement if not available
#
if {[llength [info commands lmap]] == 0} { proc lmap args {
	# 8.5 doesn't have lmap
	# https://wiki.tcl-lang.org/page/lmap+forward+compatibility
	set body [lindex $args end]
	set args [lrange $args 0 end-1]
	set n 0
	set pairs [list]
	foreach {varnames listval} $args {
		set varlist [list]
		foreach varname $varnames {
			upvar 1 $varname var$n
			lappend varlist var$n
			incr n
		}
		lappend pairs $varlist $listval
	}
	set temp [list]
	foreach {*}$pairs {
		lappend temp [uplevel 1 $body]
	}
	set temp
}}; # lmap

proc getNextOrPrevMatch {aList aString {curIdx 0} {backwards no}} {
	set idxs [lsearch -all -nocase $aList $aString*]
	if {[llength $idxs] == 0} { return -1 }
	if {[llength $idxs] == 1 && [lindex $idxs 0] == $curIdx} { return -1 }
	set p -1
	foreach idx $idxs {
		if {$idx < $curIdx} { set p $idx }
		if {$idx > $curIdx} { break }
	}
	if {$backwards} {
		if {$p == -1} {
			set newIdx [lindex $idxs end]
		} else {
			set newIdx $p
		}	
	} else {
		if {$idx == $curIdx} {
			set newIdx [lindex $idxs 0]
		} else {
			set newIdx $idx
		}
	}
	return $newIdx
}

proc makeRWAccessors {args} {
	set c ""
	foreach m $args { append c method " " $m " " args " {" \
		[string trimright [regsub -all -line {^ +} [string map [list @m@ $m] {
			if {[llength $args] == 1} { set @m@ [lindex $args 0] }
			return $@m@
		}] "\t"]] "\n}\n"
	}
	return [string trimright $c]
}
proc makeROAccessors {args} {
	set c ""
	foreach m $args { append c method " " $m " " "{}" " {" \
		[string trimright [regsub -all -line {^ +} [string map [list @m@ $m] {
			return $@m@
		}] "\t"]] "\n}\n"
	}
	return [string trimright $c]
}

proc hexdump {blob {bits 8}} {
	set l [string length $blob]
	set dump {}
	for {set i 0; set ii [expr {$bits - 1}]; set addr 0} {$i < $l} {incr i $bits; incr ii $bits; incr addr $bits} {
		binary scan [string range $blob $i $ii] @0H*@0a* hex ascii
		regsub -all -- {[^[:graph:] ]} $ascii {.} ascii
		regsub -all -- {..} $hex {& } hex
		lappend dump [format "%04X  %-[expr {3 * $bits}]s %-${bits}s" $addr $hex $ascii]
	}
	return $dump
}

#
# Remove blank lines and comment lines, add space-backslash to all lines but the last.
# If a line contains ";#" (spaces ok too) then chares from those to the end of the line will be removed
# Useful for making big lists in code without all the unpleasant backslashes.
#
proc backslashify {blob} { return [join [lmap l [split $blob \n] { if {[regexp {^[[:space:]]*(;?#.*)?$} $l]} { continue } else { lindex [regexp -inline -- {^(.+?)(;\s*#.*)?$} $l] 1 } }] " \\\n"] }

proc maxstrlen {args} { return [::tcl::mathfunc::max 0 {*}[lmap s $args { string length $s }]] }

proc oneOf {l} {
	if {[llength $l] < 2} { return [join $l] }
	return [strcat "one of " [join [lrange $l 0 end-1] ", "] [expr {[llength $l] > 2 ? "," : ""}] " or " [lindex $l end]]
}


# File
namespace eval file {
proc getFileFor {what {fn ""} args} {
	if {$fn ni {"" "?"}} { return $fn }
	if {[package provide Tk] eq ""} { return "" }
	return [${what}FileDialog {*}$args]
}
proc getLoadFile {{fn ""} args} {
	dict lappend args -filetypes [list "All Files" [list *]]
	return [getFileFor load $fn {*}$args]
}
proc getSaveFile {{fn ""} args} {
	dict lappend args -filetypes [list "All Files" [list *]]
	return [getFileFor save $fn {*}$args]
}
proc openFile {fn {mode r}} {
	if {$fn eq "-"} { return [expr {$mode eq "r" ? "stdin" : "stdout"}] }
	if {$fn in {stdin stdout stderr}} { return $fn }
	if {[catch {open $fn $mode} f]} { return "" }
	return $f
}
proc closeFile {f} {
	if {$f in {stdin stdout stderr}} { return }
	close $f
}
proc loadFile {{fn ""} args} {
	set fn [getLoadFile $fn {*}$args]
	if {$fn eq ""} { return [list "" "" 0] }
	if {[set f [openFile $fn]] eq ""} { return [list "" "" 0] }
	set data [read -nonewline $f]
	closeFile $f
	return [list $fn $data 1]
}
proc saveFile {fn data args} {
	set fn [getSaveFile $fn {*}$args]
	if {$fn eq ""} { return [list "" 0] }
	if {[set f [openFile $fn w]] eq ""} { return [list "" "" 0] }
	puts -nonewline $f $data
	closeFile $f
	return [list $fn 1]
}
proc loadListFile {{fn ""} args} {
	if {![lassign [loadFile $fn {*}$args] fn data]} { return [list $fn "" 0] }
	return [list $fn [processListFileBlob $data] 1]
}
proc processListFileBlob {blob} {
	set res [list]
	foreach l [split $blob \n] {
		set l [string trim $l]
		if {$l eq {} || [string index $l 0] eq {#}} { continue }
		if {[catch {set len [llength $l]}] || $len < 1} { continue }
		lappend res $l
	}
	return $res
}
proc filemode->text {mode} {
    foreach { mask pairs } {
	00400 { 00400 r }
	00200 { 00200 w }
	04100 { 04100 s 04000 S 00100 x }
	00040 { 00040 r }
	00020 { 00020 w }
	02010 { 02010 s 02000 S 00010 x }
	00004 { 00004 r }
	00002 { 00002 w }
	01001 { 01001 t 01000 T 00001 x }
    } {
	set value [expr $mode & $mask]
	set bit -
	foreach { x b } $pairs {
	    if { $value == $x } {
		set bit $b
	    }
	}
	append retval $bit
    }
    return $retval
}
# Tk

proc loadFileDialog {args} {
	if {$::tcl_platform(platform) ne "unix" || [catch {package require fsdialog}]} {
		tailcall tk_getOpenFile {*}$args
	} else {
		tailcall ttk::getOpenFile {*}$args
	}
}
proc saveFileDialog {args} {
	if {$::tcl_platform(platform) ne "unix" || [catch {package require fsdialog}]} {
		tailcall tk_getSaveFile {*}$args
	} else {
		tailcall ttk::getSaveFile {*}$args
	}
}
proc chooseDirectoryDialog {args} {
	if {$::tcl_platform(platform) ne "unix" || [catch {package require fsdialog}]} {
		tailcall tk_chooseDirectory {*}$args
	} else {
		tailcall ttk::chooseDirectory {*}$args
	}
}
proc loadFileIntoText {w {fn ""} args} {
	lassign [loadFile $fn {*}$args] fn data
	if {$fn eq ""} { return "" }
	$w delete 1.0 end
	$w insert end $data
	return $fn
}
proc saveSelectionToFile {w {fn ""} args} {
	if {[llength [set sel [$w tag ranges sel]]] != 2} { return }
	set blob [string trim [$w get {*}$sel]]
	if {$blob eq ""} { return }
	saveFile $fn $blob {*}$args
	return $fn
}
namespace export openFile closeFile getLoadFile loadFile getSaveFile saveFile
namespace export loadListFile processListFileBlob
namespace export loadFileDialog saveFileDialog
namespace export loadFileIntoText saveSelectionToFile
namespace export filemode->text
}; #namespace eval file



# Tk

proc bindings {tag} {
	set d [dict create]
	set l [expr {[string index $tag 0] eq "." ? [bindtags $tag] : [list $tag]}]
	foreach t $l { foreach b [bind $t] {
		dict set d $t $b [bind $t $b]
	}}
	return $d
}
proc prettyBindings {bindings} {
	set q ""
	dict for {k v} $bindings {
		append q $k \n
		dict for {b s} $v {
			append q "  " [format "%-30s" $b] " " $s \n
		}
	}
	return $q
}


# Windows and toplevels

proc randWindow {{w {.}}} {
	while {[winfo exists [set rw $w[randstr]]]} {}
	return $rw
}
proc windowIsToplevel {w} {
#	return [expr {![string match .?*.?* $w]}]
#	return [expr {[winfo toplevel $w] eq $w}]
	return [expr {[winfo toplevel $w] eq $w || [winfo class $w] eq "Toplevel"}]
}
proc newToplevel {{w ""} args} {
	if {$w eq ""} { set w [randWindow] }
	toplevel $w {*}$args
	return $w
}
proc groupToplevel {w {W ""}} {
	variable defaultWindow
	if {$W eq ""} { set W $defaultWindow }
	wm group $w $W
}
proc destroyToplevel {w} {
	destroy $w
}
proc transientizeToplevel {w {W ""}} {
	variable defaultWindow
	if {$W eq ""} { set W $defaultWindow }
	wm transient $w $W
}
proc withdrawToplevel {w} {
	wm withdraw $w
}
proc forgetToplevel {w} {
	wm forget $w
}
proc deiconifyToplevel {w} {
	wm deiconify $w
}
proc toplevelIsZoomable {w} {
	return [expr {"-zoomed" in [dict keys [wm attributes $w]]}]
}
proc toplevelZoomed {w {zoomed {}}} {
	if {$zoomed ne {}} { wm attributes $w -zoomed $zoomed }
	return [wm attributes $w -zoomed]
}
proc toplevelFullscreen {w {fullscreen {}}} {
	if {$fullscreen ne {}} { wm attributes $w -fullscreen $fullscreen }
	return [wm attributes $w -fullscreen]
}
proc toplevelGeometry {w args} {
	if {[llength $args] == 1} { wm geometry $w [lindex $args 0] }
	return [wm geometry $w]
}
proc manageToplevel {w} {
	wm manage $w
}
proc toplevelResizable {w args} {
	if {[llength $args] == 2} { wm resizable $w {*}$args }
	return [wm resizable $w]
}
proc setToplevelTitle {w t} {
	wm title $w $t
}
proc getToplevelTitle {w} {
	return [wm title $w]
}
proc centreToplevel {w {W ""}} {
	variable defaultWindow
	if {$W eq ""} { set W $defaultWindow }
	::tk::PlaceWindow $w widget $W
}
proc onToplevelDelete {w s} {
	if {[winfo toplevel $w] eq $w} {
		wm protocol $w WM_DELETE_WINDOW $s
	} else {
		bind $w <Destroy> $s
	}
}
proc createToplevelGui {{title ""} args} {
	set w [newToplevel {} {*}$args]
	groupToplevel $w
	if {$title ne ""} { setToplevelTitle $w $title }
	return $w
}

proc comboGo {combo where} {
	if {[set curIdx [$combo current]] == {}} { return }
	set nitems [llength [$combo cget -values]]
	incr nitems -1
	set idx $curIdx
	switch -exact -- $where {
	home    {  set idx 0       }
	end     {  set idx $nitems }
	next    { incr idx         }
	prior   { incr idx -1      }
	default {  set idx $where  }
	}; # hctiws
	if {$idx < 0 || $idx > $nitems || $idx == $curIdx} { return }
	$combo current $idx
	event generate $combo <<SelectAll>>
	event generate $combo <<ComboboxSelected>>
}
proc comboKey {combo key} {
	if {$key eq ""} { return }
	if {[set curIdx [$combo current]] == {}} { return }
	if {[set newIdx [getNextOrPrevMatch [$combo cget -values] $key $curIdx [string is upper $key]]] == -1} { return }
	comboGo $combo $newIdx
}
proc combolistGo {combolist where} {
	if {[set curIdx [lindex [$combolist curselection] 0]] == {}} { return }
	if {$where eq "home"} {
		set idx 0
	} elseif {$where eq "end"} {
		set idx [$combolist index end]
		incr idx -1
	} else {
		return
	}
	if {$idx == $curIdx} { return }
	$combolist activate $idx
	$combolist selection clear $curIdx
	$combolist selection set $idx
	$combolist see $idx
}
proc combolistKey {combolist key} {
	if {[string length $key] != 1} { return }
	if {[set curIdx [lindex [$combolist curselection] 0]] == {}} { return }
	# XXX Very hacky for Ttk. Who knows if/when this may change?
	if {[set newIdx [getNextOrPrevMatch [[winfo parent [winfo parent [winfo parent $combolist]]] cget -values] $key $curIdx [string is upper $key]]] == -1} { return }
	$combolist activate $newIdx
	$combolist selection clear $curIdx
	$combolist selection set $newIdx
	$combolist see $newIdx
}

proc setEntryText {w text} {
	$w delete 0 end
	if {$text ne ""} { $w insert end $text }
}

proc setTextText {w text} {
	if {[set state [$w cget -state]] ne "normal"} { $w configure -state normal }
	$w delete 1.0 end
	if {$text ne ""} { $w insert end $text }
	#if {$text ne ""} { $w insert end $text[expr {[string index $text end] eq "\n" ? "" : "\n"}] }
	if {$state ne "normal"} { $w configure -state $state }
}

# From the Tk widget demo "A multi-column list of countries"
# Code to do the sorting of the tree contents when clicked on
#
proc tvSortBy {tree col direction} {
	# Build something we can sort
	set data {}
	foreach row [$tree children {}] {
		lappend data [list [$tree set $row $col] $row]
	}

	set dir [expr {$direction ? "-decreasing" : "-increasing"}]
	set r -1

	# Now reshuffle the rows into the sorted order
	foreach info [lsort -dictionary -index 0 $dir $data] {
		$tree move [lindex $info 1] {} [incr r]
	}

	# Switch the heading so that it will sort in the opposite direction
	$tree heading $col -command [list [namespace current]::tvSortBy $tree $col [expr {!$direction}]]
}

#
# Underline Ampersand
# Convert ampersand in text to -underline option
# From Gub
#
proc ua {text} {	
	set index 0
	set uli -1
	while {[set index [string first "&" $text $index]] != -1} {
		if {[string index $text $index+1] eq "&"} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			incr index
			continue
		}
		if {$index < [string length $text]-1} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			set uli $index
		}
		incr index
	}
	set res [list $text]
	if {$uli != -1} {
		lappend res -underline $uli
	}
	return $res
}

# Edit (popup|program) menu & procs
#
# Keeps track of two menus that can be used by a program
# An epm menu with items for choosing files etc., and edit items.
# An epm edit menu with only edit items.

namespace eval epm {
	variable M {} ;# The epm menu
	variable Mmap [dict create] ;# Named items and their indexes
	variable E {} ;# The epm edit menu
	variable Emap [dict create] ;# Named items and their indexes
	# The 'mu' for E
	variable epmmu [dict create -win {} -mmap {} -defs {} -ftmap {} -inserter {} -getwrap {} -setwrap {}]
	namespace path [linsert [namespace path] end [namespace parent]]
	variable epmww ;# Current word wrap state

# Create an epm menu
proc epmCreateMenu {{w ""} args} {
	variable M
	variable Mmap
	set M [menu $w.epm -tearoff 0 -postcommand [namespace current]::epmAdjustMenu {*}$args]
	foreach item [list "Choose &File" "Choose &Directory" "Choose C&olour" "&Set To Default" Cu&t &Copy &Paste C&lear "Select &All"] {
		$M add command -command [namespace current]::epm[string map {{ } {}} [string map {"&" ""} $item]] -label {*}[ua $item]
	}
	$M add checkbutton -command [namespace current]::epmAdjustWordWrap -variable [namespace current]::epmww -onvalue word -offvalue none -label {*}[ua "&Word Wrap"]
	$M insert 4 separator
	set Mmap [dict create cut 5 copy 6 paste 7 clear 8 wrap 10 setdef 3]
	return $M
}
# Return the epm menu, create if doesn't exist
proc epmMenu {} {
	variable M
	if {$M eq {}} { epmCreateMenu }
	return $M
}
# Create an edit menu
proc epmCreateAnEditMenu {{w ""} args} {
	set e [menu $w.eepm_[randstr] -tearoff 0 {*}$args]
	foreach item [list Cu&t &Copy &Paste C&lear "Select &All" "Sa&ve As"] {
		$e add command -command [namespace current]::epm[string map {{ } {}} [string map {"&" ""} $item]] -label {*}[ua $item]
	}
	$e add checkbutton -command [namespace current]::epmAdjustWordWrap -variable [namespace current]::epmww -onvalue word -offvalue none -label {*}[ua "&Word Wrap"]
	set emap [dict create cut 0 copy 1 paste 2 clear 3 saveas 5 wrap 6]
	epmSetEpmAdjustMenu $e $emap
	return [list $e $emap]
}
# Setup epm adjustment for a menu
proc epmSetEpmAdjustMenu {m mmap} {
	$m configure -postcommand [list [namespace current]::epmAdjustMenu $m $mmap]
}
# Create the epm edit menu
proc epmCreateEditMenu {{w ""}} {
	variable E
	variable Emap
	lassign [epmCreateAnEditMenu $w] E Emap
	return $E
}
# Return the epm edit menu, create if doesn't exist
proc epmEditMenu {} {
	variable E
	if {$E eq {}} { epmCreateEditMenu }
	return $E
}
# Have the epm menu popup for a widget
proc epmBindEpmMenuTo {w args} {
	variable Mmap
	# Call epmMenu to ensure M's menu exists
	tailcall epmBindMenuTo $w [epmMenu] {*}$args -mmap $Mmap
}
# Have the epm edit menu popup for a widget
proc epmBindEditMenuTo {w args} {
	variable Emap
	# Call epmEditMenu to ensure E's menu exists
	tailcall epmBindMenuTo $w [epmEditMenu] {*}$args -mmap $Emap
}
# Make a 'mu'
proc epmMakeMu {m args} {
	set ns [namespace current]:
	set mu [dict merge $args [dict create -menu $m \
		-mmap     [dgwd [dict create]      $args -mmap]     \
		-defs     [dgwd [dict create]      $args -defs]     \
		-ftmap    [dgwd [dict create]      $args -ftmap]    \
		-inserter [dgwd $ns:epmInserter    $args -inserter] \
		-getwrap  [dgwd $ns:epmGetWordWrap $args -getwrap]  \
		-setwrap  [dgwd $ns:epmSetWordWrap $args -setwrap]  \
	]]
	return $mu
}
# Have a menu popup for a widget
proc epmBindMenuTo {w m args} {
	set mu [epmMakeMu $m {*}$args]
	set mod   [dgwd ""        $args -mod]
	set seq   [dgwd "Button-" $args -seq]
	set bn    [dgwd 3         $args -bn]
	bind $w <$mod$seq$bn> [list [namespace current]::epmMenuGo %X %Y %W $mu]
}
# Popup activator
proc epmMenuGo {x y w mu} {
	variable epmmu [dict merge $mu [dict create -win $w -x $x -y $y]]
	tk_popup [dict get $epmmu -menu] $x $y
}
# Return the 'mu'
proc epmMu {} {
	variable epmmu
	return $epmmu
}
# Current window for a popup
proc epmW {args} {
	variable epmmu
	if {[llength $args] == 1} {
		dict set epmmu -win [lindex $args 0]
	}
	return [dict get $epmmu -win]
}
# Get the word wrap state for a window. Used by the default 'mu'.
proc epmGetWordWrap {w} {
	return [$w cget -wrap]
}
# Get the word wrap state for a window. Used by the default 'mu'.
proc epmSetWordWrap {w wrap} {
	$w configure -wrap $wrap
}
# Adjust a window's word wrap state according to epm state
proc epmAdjustWordWrap {} {
	variable epmmu
	variable epmww
	[dict get $epmmu -setwrap] [dict get $epmmu -win] $epmww
}
# Ajdust a menu according to a window's state
proc epmAdjustMenu {{m {}} {mmap {}}} {
	variable epmmu
	variable epmww
	set w [dict get $epmmu -win]
	if {$w eq "" || ![winfo exists $w]} {
		if {![catch {info level -1} res]} {
			::tsm::log "epmAdjustMenu !$res!" error
		}
		return
	}
	set defs [dict get $epmmu -defs]
	if {$m eq {}} {
		set m [dict get $epmmu -menu]
	}
	if {$mmap eq {}} {
		set mmap [dict get $epmmu -mmap]
	}
	set items [list]
	if {[dict size $mmap] > 0} { foreach what [list cut copy paste clear wrap setdef saveas] {
		if {![dict exists $mmap $what]} { continue }
		lappend items $what
	}}
	if {[llength $items] == 0} { return }
	set states [dict create]
	dict set states setdef [expr {[dict exists $defs $w] ? "normal" : "disabled"}]
	set wclass [winfo class $w]
	if {$wclass eq "Text"} {
		set epmww [[dict get $epmmu -getwrap] $w]
		if {$epmww eq "char"} { set epmww word }
		dict set states wrap normal
		set sel [expr {[llength [$w tag ranges sel]] > 0 ? "normal" : "disabled"}]
	} elseif {$wclass eq "TLabel"} {
		set epmww none
		dict set states wrap disabled
		set sel normal
	} else {
		set epmww none
		dict set states wrap disabled
		set sel [expr {[$w selection present] ? "normal" : "disabled"}]
	}
	dict set states copy $sel
	dict set states saveas $sel
	if {[$w cget -state] eq "normal"} {
		foreach item [list cut clear] {
			dict set states $item $sel
		}
		dict set states paste normal
	} else {
		foreach item [list cut clear paste setdef saeveas] {
			dict set states $item disabled
		}
	}
	if {[dict get $states paste] eq "normal" && "paste" in $items} {
		if {[catch {selection get -displayof $w -selection CLIPBOARD} s]} { set s "" }
		dict set states paste [expr {[string length $s] > 0 ? "normal" : "disabled"}]
	}
	foreach item $items {
		$m entryconfigure [dict get $mmap $item] -state [dict get $states $item]
	}
}
proc epmInserter {where what} {
	epmSelectAll ; epmClear
	$where insert end $what
}
proc epmChooseFile {} {
	variable epmmu
	set w [dict get $epmmu -win]
	set ftmap [dict get $epmmu -ftmap]
	set filetypes [dgwd {} $ftmap $w]
	if {[lsearch -exact -index {0 1} $filetypes [list *]] == -1} {
		lappend filetypes [list "All Files" [list *]]
	}
	if {[set v [::tu::file::loadFileDialog -initialdir [file dirname [$w get]] -initialfile [$w get] -filetypes $filetypes]] eq ""} { return }
	[dict get $epmmu -inserter] $w $v
}
proc epmChooseDirectory {} {
	variable epmmu
	set w [dict get $epmmu -win]
	if {[set v [::tu::file::chooseDirectoryDialog -initialdir [$w get]]] eq ""} { return }
	[dict get $epmmu -inserter] $w $v
}
proc epmChooseColour {} {
	variable epmmu
	set w [dict get $epmmu -win]
	if {[winfo class $w] eq "Text"} {
		set v ""
	} else {
		set v [string trim [$w get]]
	}
	if {$v eq ""} { set v #000000 }
	if {[catch {winfo rgb . $v}]} { set v olive }
	if {[set v [tk_chooseColor -initialcolor $v]] eq ""} { return }
	[dict get $epmmu -inserter] $w $v
}
proc epmSaveAs {} {
	variable epmmu
	::tu::file::saveSelectionToFile [dict get $epmmu -win]
}
proc epmSetToDefault {} {
	variable epmmu
	set w [dict get $epmmu -win]
	set defs [dict get $epmmu -defs]
	if {![dict exists $defs $w]} { return }
	set v [dict get $defs $w]
	[dict get $epmmu -inserter] $w $v
}
foreach e [list Cut Copy Paste Clear SelectAll] {
	proc epm$e {} [string map [list @e@ $e] {
		variable epmmu
		set w [dict get $epmmu -win]
		event generate $w <<@e@>>
		focus $w
	}]
}; unset e

namespace export epmCreateMenu epmMenu
namespace export epmCreateEditMenu epmCreateAnEditMenu epmEditMenu
namespace export epmBindMenuTo epmBindEpmMenuTo epmBindEditMenuTo
namespace export epmMakeMu epmSetEpmAdjustMenu
namespace export epmMu epmW

}; #namespace eval epm

if {[llength [info commands ::oo::class]] == 1} { ::oo::class create MyFonts {
	variable TkFonts
	variable class
	variable fonts
	variable defaults
constructor {myClass} {
	set class $myClass
	set TkFonts [list Default Text Fixed Menu Heading]
	set fonts [dict create]
	my setup
}
eval [::tu::makeROAccessors class TkFonts fonts]
method setup {} {
	foreach f $TkFonts {
		set c [font configure Tk${f}Font]
		dict set fonts [font create ${class}${f}Font {*}$c] {}
	}
	foreach {attr which props} [list {*}{
		Bigger Text {-size 11}
		Log    Text {}
	}] {
		dict set fonts [font create ${class}${attr}${which}Font {*}[font configure ${class}${which}Font] {*}$props] {}
	}
#		Bold   Text {-weight bold}
#		Bold Fixed   {-weight bold}
#		Big  Default {-size -18}

	foreach f [dict keys $fonts] { dict set fonts $f [font configure $f] }
	option add *${class}*Menu*font      ${class}MenuFont
	option add *${class}*Text*font      ${class}FixedFont
	option add *${class}*TLabel*font    ${class}DefaultFont
	option add *${class}*TEntry*font    ${class}TextFont
	option add *${class}*TCombobox*font ${class}TextFont

	ttk::style configure Heading           -font ${class}HeadingFont
	ttk::style configure Treeview          -font ${class}DefaultFont
	ttk::style configure TButton           -font ${class}DefaultFont
	ttk::style configure TCheckbutton      -font ${class}DefaultFont
	ttk::style configure TRadiobutton      -font ${class}DefaultFont
	ttk::style configure TNotebook.Tab     -font ${class}DefaultFont
	ttk::style configure TLabelframe.Label -font ${class}DefaultFont
}
method defaults {font} {
	return [dict get $fonts $font]
}
method reset {} {
	dict for {f c} $fonts { font configure $f {*}$c }
}
method transform-scale {{s 0} args} {
	if {[llength $args] > 0} {
		set fnts [dict create]
		foreach arg $args { dict set fnts $arg [dict get $fonts $arg] }
	} else {
		set fnts $fonts
	}
	dict for {f c} $fnts {
		set size [dict get $c -size]
		font configure $f -size [expr {$size + (($size < 0 ? -1 : 1) * $s)}]
	}
}
}}; # ::oo::class create MyFonts if ::oo::class exists


namespace import epm::*
namespace export {*}[namespace eval epm { namespace export }]

namespace import file::*
namespace export {*}[namespace eval file { namespace export }]

namespace export \
	randWindow windowIsToplevel newToplevel groupToplevel destroyToplevel \
	transientizeToplevel centreToplevel \
	withdrawToplevel forgetToplevel deiconifyToplevel \
	onToplevelDelete setToplevelTitle getToplevelTitle \
	toplevelIsZoomable toplevelZoomed toplevelFullscreen \
	toplevelGeometry manageToplevel toplevelResizable \
namespace export randstr dgwd imn hexdump backslashify maxstrlen oneOf
namespace export makeRWAccessors makeRWOccessors
namespace export tvSortBy
namespace export comboGo comboKey combolistGo combolistKey
namespace export setEntryText setTextText
namespace export createToplevelGui
namespace export ua
namespace export bindings prettyBindings

package provide tu 0.2

}; # namespace eval tu

# EOF
